@@ -12,7 +12,7 @@ import zipfile
 logger = logging.getLogger(__name__)
 
 
-def copytree(src, dst, symlinks=0, allowed_basenames=None):
+def copytree(src, dst, symlinks=0, allowed_basenames=None, exclude=[]):
     """Local implementation of shutil's copytree function.
 
     Checks wheather destination directory exists or not
@@ -32,12 +32,14 @@ def copytree(src, dst, symlinks=0, allowed_basenames=None):
                 if name not in allowed_basenames:
                     logger.debug("Skipped %s" % srcname) 
                     continue
+        if os.path.basename(srcname) in exclude:
+            continue
         try:
             if symlinks and os.path.islink(srcname):
                 linkto = os.readlink(srcname)
                 os.symlink(linkto, dstname)
             elif os.path.isdir(srcname):
-                copytree(srcname, dstname, symlinks, allowed_basenames)
+                copytree(srcname, dstname, symlinks, allowed_basenames, exclude)
             elif not os.path.isfile(dstname) and symlinks:
                 os.symlink(srcname, dstname)
             elif not symlinks:
@@ -84,8 +86,8 @@ class Recipe(zc.recipe.egg.Eggs):
 
         super(Recipe, self).__init__(buildout, name, opts)
         opts['app-directory'] = os.path.join(buildout['buildout']
-                                              ['parts-directory'],
-                                              self.name)
+                                             ['parts-directory'],
+                                             self.name)
         opts['bin-directory'] = buildout['buildout']['bin-directory']
 
     def write_server_script(self, name, bin, sys_path):
@@ -142,7 +144,6 @@ class Recipe(zc.recipe.egg.Eggs):
         script.close()
         os.chmod(path, 0755)
 
-
     def install_appengine(self):
         """Downloads and installs Google App Engine."""
         arch_filename = self.options['url'].split(os.sep)[-1]
@@ -183,10 +184,20 @@ class Recipe(zc.recipe.egg.Eggs):
         gae_config = os.path.join(gae, 'appcfg.py')
         self.write_appcfg_script(gae_config, sys_path)
 
+    def write_pkg_resources_stub(self, d):
+        """Writes a stub for setuptool's pkg_resources module."""
+        pkg_resources_stub = open(os.path.join(d, 'pkg_resources.py'), "w")
+        pkg_resources_stub.write("def _dummy_func(*args):\n")
+        pkg_resources_stub.write("    pass\n\n")
+        pkg_resources_stub.write("declare_namespace = _dummy_func\n")
+        pkg_resources_stub.write("resource_filename = _dummy_func\n")
+        pkg_resources_stub.close()
+
     def copy_packages(self, ws, lib):
         """Copy egg contents to lib-directory."""
         if not os.path.exists(lib):
             os.mkdir(lib)
+        self.write_pkg_resources_stub(lib)
         packages = self.options.get('packages', '').split()
         keys = [k.lower() for k in packages]
         for p in keys:
@@ -208,7 +219,9 @@ class Recipe(zc.recipe.egg.Eggs):
             sources.close()
             if not os.path.exists(dir):
                 os.mkdir(dir)
-            copytree(src, dir, True, allowed_basenames=allowed_basenames)
+            copytree(src, dir, True,
+                     allowed_basenames=allowed_basenames,
+                     exclude=self.options.get('exclude', '').split())
 
     def zip_packages(self, ws, lib):
         """Creates zip archive of configured packages."""
@@ -219,16 +232,7 @@ class Recipe(zc.recipe.egg.Eggs):
         os.chdir(lib)
         for root, dirs, files in os.walk('.'):
             for f in files:
-                ext = os.path.splitext(f)[1]
-                if ext not in ('.h', '.c'):
-                    zipper.add(os.path.join(root, f))
-        pkg_resources_dummy = open(os.path.join(lib, 'pkg_resources.py'), "w")
-        pkg_resources_dummy.write("def _dummy_func(*args):\n")
-        pkg_resources_dummy.write("    pass\n\n")
-        pkg_resources_dummy.write("declare_namespace = _dummy_func\n")
-        pkg_resources_dummy.write("resource_filename = _dummy_func\n")
-        pkg_resources_dummy.close()
-        zipper.add('pkg_resources.py')
+                zipper.add(os.path.join(root, f))
         zipper.close()
 
     def copy_sources(self):
@@ -244,7 +248,8 @@ class Recipe(zc.recipe.egg.Eggs):
             reqs, ws = self.working_set()
             sources = [d.location for d in ws if d.key in reqs]
         for s in sources:
-            copytree(s, options['app-directory'], True)
+            copytree(s, options['app-directory'], True,
+                     exclude=options.get('exclude', '').split())
 
     def install(self):
         """Creates the part."""
@@ -253,16 +258,20 @@ class Recipe(zc.recipe.egg.Eggs):
         if not self.options.get('appengine-lib', False):
             self.install_appengine()
         self.setup_bin(ws)
-        temp_dir = os.path.join(tempfile.mkdtemp(), self.name)
-        self.copy_packages(ws, temp_dir) 
         app_dir = options['app-directory']
+        if options.get('zip-packages', 'YES').lower() in ['yes', 'true']:
+            temp_dir = os.path.join(tempfile.mkdtemp(), self.name)
+        else:
+            temp_dir = app_dir
         if os.path.isdir(app_dir):
             shutil.rmtree(app_dir, True)
         if not os.path.exists(app_dir):
             os.mkdir(app_dir)
-        self.zip_packages(ws, temp_dir)
-        if os.path.isdir(temp_dir):
-            shutil.rmtree(temp_dir, True)
+        self.copy_packages(ws, temp_dir) 
+        if options.get('zip-packages', 'YES').lower() in ['yes', 'true']:
+            self.zip_packages(ws, temp_dir)
+            if os.path.isdir(temp_dir):
+                shutil.rmtree(temp_dir, True)
         self.copy_sources()
         return ()
 
